//
//  TestController.m
//  ImageTest
//
//  Created by Sebastian Hans on Fri Aug 27 2004.
//  Copyright (c) 2004 Sebastian Hans. All rights reserved.
//

#import "TestController.h"
#import "Generic8BitImage.h"
#import <stdio.h>

@implementation TestController

- (void) awakeFromNib
{
	[super awakeFromNib];
	model_list = cp_create_model_list([numPointsField intValue]);
	match_list = NULL;
	current_model = NULL;
	next_model_id = 1;
	current_difference_stepper_value = 0;
}

- (IBAction) processImage: (id) sender
{
	[self rematch: sender];
}

- (IBAction) rematch: (id) sender
{
	NSImage *img;
	Generic8BitImage *generic_img;
	IplImage *ipl_img;
	cp_match_t *best_match, *match;

	if (match_list)
	{
		[differenceStepper setEnabled: NO];
		cp_destroy_match_list(match_list);
		match_list = NULL;
	}
	if (!model_list->length)
	{
		puts("No models in list. Add models by dragging them into the left window.\n");
		return;
	}

	if (!(img = [imageView image]))
	{
		puts("No image to match.");
		return;
	}

	generic_img = [[Generic8BitImage alloc] fromNSImage: img];
	ipl_img = [generic_img toIplImage];
	[generic_img release];
	match_list = cp_match(ipl_img, model_list, &best_match);
	cvReleaseImage(&ipl_img);

	if (!match_list)
	{
		puts("Got no match list. 'Sup?");
		return;
	}
	if (!match_list->length)
	{
		puts("Empty match list.");
		return;
	}

	cp_sort_match_list(match_list);

	for (match = match_list->first; match; match = match->next)
	{
		if (match->model->name)
			printf("Model %ld (%s) matches with %g, rel. orient: %g, rel. size: %g\n",
					match->model->id, match->model->name, match->difference,
					match->relative_orientation, match->relative_size);
		else
			printf("Model %ld matches with %g, rel. orient: %g, rel. size: %g\n",
					match->model->id, match->difference, match->relative_orientation,
					match->relative_size);
		if (!best_match || match->difference < best_match->difference)
			best_match = match;
	}
	[self setCurrentModel: best_match->model];
	[differenceStepper setIntValue: current_difference_stepper_value];
	[differenceStepper setEnabled: YES];
}

- (void) setCurrentModel: (cp_model_t*) model
{
	Generic8BitImage *generic_img;
	NSString *name;

	current_model = model;
	if (model)
	{
		generic_img = [[Generic8BitImage alloc] fromIplImage: model->image];
		[modelView setImage: [generic_img toNSImage]];
		[generic_img release];
		name = [self cString2NSString: model->name];
		[modelComboBox setStringValue: name];
		[modelComboBox selectItemWithObjectValue: name];
	}
	else
	{
		[modelView setImage: nil];
		[modelComboBox setStringValue: @""];
	}
	[self updateDifference];
}

- (void) updateDifference
{
	cp_match_t *match = NULL;
	char diff_buf[12];

	if (match_list)
		match = cp_find_match_for_model(match_list, current_model);
	if (match)
	{
		snprintf(diff_buf, 12, "%6.3f", match->difference);
		diff_buf[11] = '\0';
		[differenceField setTitleWithMnemonic: [self cString2NSString: diff_buf]];
	}
	else
		[differenceField setTitleWithMnemonic: [self cString2NSString: "-"]];
}

- (NSString*) cString2NSString: (char*) s
{
	return [[NSString alloc] initWithCStringNoCopy: s length: strlen(s) freeWhenDone: NO];
}

- (IBAction) addModel: (id) sender
{
	Generic8BitImage *img;
	char name_buf[256], *name;

	img = [[Generic8BitImage alloc] fromNSImage: [sender image]];
	snprintf(name_buf, 256, "Model %ld", next_model_id);
	name_buf[255] = '\0';
	if (!(name = malloc(strlen(name_buf) + 1)))
	{
		puts("No memory for the model name?");
		return;
	}
	memcpy(name, name_buf, strlen(name_buf) + 1);
	printf("Adding model '%s'.\n", name);
	cp_append_model(model_list, [img toIplImage], next_model_id, name);
	[img release];
	[modelComboBox addItemWithObjectValue: [self cString2NSString: name]];
	if ([imageView image])
		[self rematch: self];
	else
		[self setCurrentModel: cp_find_model_name(model_list, name)];
	next_model_id++;
}

- (IBAction) clearImage: (id) sender
{
	if (match_list)
	{
		[differenceStepper setEnabled: NO];
		cp_destroy_match_list(match_list);
		match_list = NULL;
	}
	[imageView setImage: nil];
	[self updateDifference];
}

- (IBAction) enterModel: (id) sender
{
	char *name;
	cp_model_t *model;
	int i;

	if (!model_list->length)
		return;

	name = [[modelComboBox stringValue] cString];
	if (!strcmp("", name))
	{
		[self setCurrentModel: current_model];
		return;
	}
	if (model = cp_find_model_name(model_list, name))
	{
		[self setCurrentModel: model];
		return;
	}
	i = [modelComboBox indexOfItemWithObjectValue:
			[self cString2NSString: current_model->name]];
	[modelComboBox removeItemAtIndex: i];
	free(current_model->name);
	current_model->name = strdup(name);
	[modelComboBox insertItemWithObjectValue: [self cString2NSString: current_model->name]
			atIndex: i];
}

- (IBAction) removeModel: (id) sender
{
	cp_model_t *next;

	if (!model_list->length)
		return;
	if (current_model->next)
		next = current_model->next;
	else if (current_model->prev)
		next = current_model->prev;
	else
		next = NULL;
	cp_remove_model(model_list, current_model);
	[modelComboBox removeItemWithObjectValue: [self cString2NSString: current_model->name]];
	cp_destroy_model(current_model, 1, 0);
	[self setCurrentModel: next];
	if ([imageView image])
		[self rematch: self];
}

- (IBAction) useModelAsImage: (id) sender
{
	if ([modelView image])
	{
		[imageView setImage: [modelView image]];
		[self processImage: imageView];
	}
}

- (IBAction) useImageAsModel: (id) sender
{
	if ([imageView image])
	{
		[modelView setImage: [imageView image]];
		[self addModel: modelView];
	}
}

- (IBAction) stepDifference: (id) sender
{
	cp_match_t *match;
	
	if(!(match = cp_find_match_for_model(match_list, current_model)))
		return;

	if ([sender intValue] == (current_difference_stepper_value + 1) % 3)
	{
		/* step up */
		if (match == match_list->last)
			match = match_list->first;
		else
			match = match->next;
	}
	else
	{
		/* step down */
		if (match == match_list->first)
			match = match_list->last;
		else
			match = match->prev;
	}
	current_difference_stepper_value = [sender intValue];
	[self setCurrentModel: match->model];
}

- (IBAction) setNumPoints: (id) sender
{
	int num_points;
	
	num_points = [sender intValue];
	if (num_points < 2)
	{
		[sender setIntValue: model_list->num_points];
		return;
	}
	if (!cp_set_num_points_for_model_list(model_list, num_points))
	{
		puts("Something went wrong changing num_points.");
		puts("I suggest restarting the program since it is now severely fucked up.");
	}
	[self rematch: self];
}

- (void) dealloc
{
	cp_destroy_model_list(model_list, 1, 0);
	if (match_list)
		cp_destroy_match_list(match_list);
	[super dealloc];
}

@end